/* This file is part of flowz.
   Copyright (C) 2006 Sergey Poznyakoff

   Flowz is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   Flowz is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with flowz; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
   MA 02110-1301 USA */

#include "flowz.h"
#define obstack_chunk_alloc malloc
#define obstack_chunk_free free

typedef struct tree_node Tree;

struct tree_node {
     Tree *next;
     Tree *left, *right;
     time_t time_start;
     time_t time_end;
     uintmax_t octets;
     char key[1];
};     

#define nodecmp(key,b) (strcmp(key, (b)->key))
 
static Tree *alloc_node(const char *key);

static Tree *
lookup_or_insert_node(Tree *branch, const char *key)
{
     int rc = nodecmp(key, branch);

     if (rc == 0) 
	  return branch;
     
     if (rc < 0) {
	  if (branch->left)
	       branch = branch->left;
	  else 
	       return branch->left = alloc_node(key);
     } else {
	  if (branch->right)
	       branch = branch->right;
	  else
	       return branch->right = alloc_node(key);
     }
     return lookup_or_insert_node(branch, key);
}	  
	  

Tree *tree;
size_t node_count;
static struct obstack stk;
static void *mem_top; 
static Tree *first_node_ptr, *prev_node_ptr;

#define first_node() first_node_ptr
#define next_node(node) (node)->next

static Tree *
alloc_node(const char *key)
{
     Tree *node = obstack_alloc(&stk, sizeof(*node) + strlen(key));
     node_count++;
     if (!first_node_ptr)
	  first_node_ptr = node;
     if (prev_node_ptr) 
	  prev_node_ptr->next = node;
     prev_node_ptr = node;
     memset(node, 0, sizeof *node);
     strcpy(node->key, key);
     return node;
}

void
table_init()
{
     obstack_init(&stk);
     mem_top = obstack_alloc(&stk, 1);
}

size_t
table_n_entries()
{
     return node_count;
}

void
table_stat(FILE *fp)
{
     fprintf(fp, "# %lu entries - %lu bytes of memory\n",
	     node_count,  obstack_memory_used(&stk));
}


int
update(char *key, time_t time_start, time_t time_end, uintmax_t octets)
{
     Tree *tp;

     if (!tree)
	  tree = tp = alloc_node(key);
     else 
	  tp = lookup_or_insert_node(tree, key);
     if (tp->time_start == 0 || tp->time_start > time_start)
	  tp->time_start = time_start;
     if (tp->time_end < time_end)
	  tp->time_end = time_end;

     if (TYPE_MAXIMUM(off_t) - tp->octets < octets) {
	  error(0, 0, "%s:%lu: integer overflow",
		input, (unsigned long) lineno);
	  return 1;
     }
     tp->octets += octets;

     if (max_memory_usage) {
	  off_t estimate = obstack_memory_used(&stk);
	  if (estimate >= max_memory_usage) {
	       dump_table(outfile, header_str);
	       obstack_free(&stk, first_node());
	       tree = NULL;
	       first_node_ptr = prev_node_ptr = NULL;
	       header_str = NULL; /* make sure header is not printed twice */
	  }
     }

     return 0;
}

void
dump_table(FILE *fp, const char *header)
{
     if (tree) {
	  size_t i;
	  char buf[UINTMAX_STRSIZE_BOUND];
	  Tree *node;
	  
	  if (header)
	       fprintf(fp, "#:%s\n", header);
	  
	  for (node = first_node(); node; node = next_node(node)) {
	       fprintf(fp, "%s-", umaxtostr(node->time_start, buf));
	       fprintf(fp, "%s,", umaxtostr(node->time_end, buf));
	       fprintf(fp, "%s,%s\n", umaxtostr(node->octets, buf), node->key);
	  }
     }
}
